#include <stdarg.h>
#include <sys/types.h>
#include <kvm.h>
#include <sys/time.h>
#include <sys/sysctl.h>
#include <limits.h>
#include <fcntl.h>
#include <stdlib.h>
#include <sys/param.h>
#include <sys/user.h>
#include <string.h>

#include "readproc.h"


int uptime(double *up, double *idle)
{
  int mib[2];
  size_t size;
  struct timeval boottime;
  time_t now;

  mib[0] = CTL_KERN;
  mib[1] = KERN_BOOTTIME;
  size = sizeof(boottime);
  if (sysctl(mib, 2, &boottime, &size, NULL, 0) == -1) return 0;
  (void)time(&now);
  return now - boottime.tv_sec;
}


unsigned read_total_main(void)
{
  int mib[2];
  size_t mem, size;

  mib[0] = CTL_HW;
  mib[1] = HW_USERMEM;
  size = sizeof(mem);
  if (sysctl(mib, 2, &mem, &size, NULL, 0) == -1) return 0;
  return mem;
}


int tty_to_dev(char *tty)
{
  /*
   * TODO: implement...
   */
  return -1;
}


PROCTAB *openproc(int flags, ...)
{
  char errbuf[_POSIX2_LINE_MAX];
  va_list ap;
  PROCTAB *tab = malloc(sizeof(PROCTAB));
  if (tab != 0) {
    va_start(ap, flags);

    tab->flags = flags;

    if (flags & PROC_PID) tab->pids = va_arg(ap, pid_t*);
    else if (flags & PROC_TTY) tab->ttys = va_arg(ap, dev_t*);
    else if (flags & PROC_STAT) tab->stats = va_arg(ap, char*);
    else if (flags & PROC_UID) {
      tab->uids = va_arg(ap, uid_t*);
      tab->nuid = va_arg(ap, int);
    }

    if ((flags & PROC_ANYTTY) && (flags & PROC_TTY))
      tab->flags &= ~PROC_TTY;

    tab->kd = kvm_openfiles(NULL, NULL, NULL, O_RDONLY, errbuf);
    if (tab->kd == 0) return 0;

    tab->kp = kvm_getprocs(tab->kd, KERN_PROC_ALL, 0, &tab->procs);
    if (tab->kp == 0) return 0;

    va_end(ap);
  }

  return tab;
}


int is_valid(PROCTAB *tab)
{
  int i = 0;

  if (tab->flags & PROC_PID) {
    while (tab->pids[i] != 0 && tab->kp->kp_proc.p_pid != tab->pids[i]) i++;
    return (tab->pids[i] == 0) ? 0 : 1;
  }

  return 1;
}


char **strsdup(char **st)
{
  char **dup;
  int i, count = 0;

  if (st != 0) while (st[count] != 0) count++;

  dup = calloc(count+1, sizeof(char*));
  if (dup == 0) return 0;

  for (i = 0; i < count; i++) dup[i] = strdup(st[i]);

  return dup;
}


proc_t *readproc(PROCTAB *tab, proc_t *p)
{
  if (tab == 0) return 0;
  if (p == 0) p = calloc(1, sizeof(proc_t));

  while (tab->procs > 0 && !is_valid(tab)) {
    tab->procs--;
    if (tab->procs > 0) tab->kp++;
    else tab->kp = 0;
  }

  if (tab->procs == 0) return 0;

  p->pid = tab->kp->kp_proc.p_pid;
  p->ppid = tab->kp->kp_eproc.e_ppid;
  p->wchan = (unsigned long)tab->kp->kp_proc.p_wchan;
  p->tty = tab->kp->kp_eproc.e_tdev;
  p->nice = tab->kp->kp_proc.p_nice;

  switch (p->tty & 0xFF00) {
  case 0x0C00: sprintf(p->ttyc, "%d", p->tty & 0xFF); break;
  case 0x0500: sprintf(p->ttyc, "p%d", p->tty & 0xFF); break;
  default:     strcpy(p->ttyc, " ?"); break;
  }

  switch (tab->kp->kp_proc.p_stat) {
  case SIDL: p->state = 'I'; break;
  case SRUN: p->state = 'R'; break;
  case SSLEEP: p->state = 'S'; break;
  case SSTOP: p->state = 'H'; break;
  case SZOMB: p->state = 'Z'; break;
  default: p->state = '.'; break;
  }

  p->cmdline = strsdup(kvm_getargv(tab->kd, tab->kp, 1024));
  p->environ = strsdup(kvm_getenvv(tab->kd, tab->kp, 1024));

  if (p->cmdline[0] != 0) p->cmd = strdup(p->cmdline[0]);
  else p->cmd = strdup("");

  tab->procs--;
  if (tab->procs > 0) tab->kp++;
  else tab->kp = 0;

  return p;
}


void  freeproc(proc_t *p)
{
  int i;

  i = 0;
  while (p->cmdline[i] != 0) free(p->cmdline[i++]);
  free(p->cmdline);

  i = 0;
  while (p->environ[i] != 0) free(p->environ[i++]);
  free(p->environ);

  free(p->cmd);
  free(p);
}


#if 0

/*
 * The following fields still have to be assigned
 */
typedef struct {
    char
    	user[10],	/* user name corresponding to owner of process */
    	ttyc[5],	/* string representation of controlling tty device */
    int
    	uid,		/* user id */
	pgrp,		/* process group id */
	session,	/* session id */
	tpgid,		/* terminal process group id */
	priority,	/* kernel scheduling priority */
    long long
	signal,		/* mask of pending signals */
	blocked,	/* mask of blocked signals */
	sigignore,	/* mask of ignored signals */
	sigcatch;	/* mask of caught  signals */
    long
	start_time,	/* start time of process -- seconds since 1-1-70 */
	utime,		/* user-mode CPU time accumulated by process */
	stime,		/* kernel-mode CPU time accumulated by process */
	cutime,		/* cumulative utime of process and reaped children */
	cstime,		/* cumulative stime of process and reaped children */
    /* the next 7 members come from /proc/#/statm */
	size,		/* total # of pages of memory */
	resident,	/* number of resident set (non-swapped) pages (4k) */
	share,		/* number of pages of shared (mmap'd) memory */
	trs,		/* text resident set size */
	lrs,		/* shared-lib resident set size */
	drs,		/* data resident set size */
	dt;		/* dirty pages */
    unsigned
        pcpu;           /* %CPU usage (is not filled in by readproc!!!) */
    unsigned long
	vsize,		/* number of pages of virtual memory ... */
	rss,		/* resident set size from /proc/#/stat */
	rss_rlim,	/* resident set size ... ? */
	timeout,	/* ? */
	it_real_value,	/* ? */
	flags,		/* kernel flags for the process */
	min_flt,	/* number of minor page faults since process start */
	maj_flt,	/* number of major page faults since process start */
	cmin_flt,	/* cumulative min_flt of process and child processes */
	cmaj_flt,	/* cumulative maj_flt of process and child processes */
	start_code,	/* address of beginning of code segment */
	end_code,	/* address of end of code segment */
	start_stack,	/* address of the bottom of stack for the process */
	kstk_esp,	/* kernel stack pointer */
	kstk_eip,	/* kernel stack pointer */
	wchan;		/* address of kernel wait channel proc is sleeping in*/
    struct proc_s *l,	/* ptrs for building arbitrary linked structs */
                  *r;	/* (i.e. singly/doubly-linked lists and trees */
} yes;


struct kinfo_proc {
        struct  proc kp_proc;                   /* proc structure */
        struct  eproc {
                struct  proc *e_paddr;          /* address of proc */
                struct  session *e_sess;        /* session pointer */
                struct  pcred e_pcred;          /* process credentials */
                struct  ucred e_ucred;          /* current credentials */
                struct  vmspace e_vm;           /* address space */
                pid_t   e_ppid;                 /* parent process id */
                pid_t   e_pgid;                 /* process group id */
                short   e_jobc;                 /* job control counter */
                dev_t   e_tdev;                 /* controlling tty dev */
                pid_t   e_tpgid;                /* tty process group id */
                struct  session *e_tsess;       /* tty session pointer */
                char    e_wmesg[WMESGLEN+1];    /* wchan message */
                segsz_t e_xsize;                /* text size */
                short   e_xrssize;              /* text rss */
                short   e_xccount;              /* text references */
                short   e_xswrss;
                long    e_flag;
                char    e_login[roundup(MAXLOGNAME, sizeof(long))];
                long    e_spare[2];
        } kp_eproc;
};


struct  proc {
        TAILQ_ENTRY(proc) p_procq;      /* run/sleep queue. */
        LIST_ENTRY(proc) p_list;        /* List of all processes. */

        /* substructures: */
        struct  pcred *p_cred;          /* Process owner's identity. */
        struct  filedesc *p_fd;         /* Ptr to open files structure. */
        struct  pstats *p_stats;        /* Accounting/statistics (PROC ONLY). */
        struct  plimit *p_limit;        /* Process limits. */
        struct  vm_object *p_upages_obj;/* Upages object */
        struct  sigacts *p_sigacts;     /* Signal actions, state (PROC ONLY). */

#define p_ucred         p_cred->pc_ucred
#define p_rlimit        p_limit->pl_rlimit

        int     p_flag;                 /* P_* flags. */
        char    p_stat;                 /* S* process status. */
        char    p_pad1[3];

        pid_t   p_pid;                  /* Process identifier. */
        LIST_ENTRY(proc) p_hash;        /* Hash chain. */
        LIST_ENTRY(proc) p_pglist;      /* List of processes in pgrp. */
        struct  proc *p_pptr;           /* Pointer to parent process. */
        LIST_ENTRY(proc) p_sibling;     /* List of sibling processes. */
        LIST_HEAD(, proc) p_children;   /* Pointer to list of children. */

        struct callout_handle p_ithandle; /*
                                              * Callout handle for scheduling
                                              * p_realtimer.
                                              */
/* The following fields are all zeroed upon creation in fork. */
#define p_startzero     p_oppid

        pid_t   p_oppid;         /* Save parent pid during ptrace. XXX */
        int     p_dupfd;         /* Sideways return value from fdopen. XXX */

        struct  vmspace *p_vmspace;     /* Address space. */

        /* scheduling */
        u_int   p_estcpu;        /* Time averaged value of p_cpticks. */
        int     p_cpticks;       /* Ticks of cpu time. */
        fixpt_t p_pctcpu;        /* %cpu for this process during p_swtime */
        void    *p_wchan;        /* Sleep address. */
        const char *p_wmesg;     /* Reason for sleep. */
        u_int   p_swtime;        /* Time swapped in or out. */
        u_int   p_slptime;       /* Time since last blocked. */

        struct  itimerval p_realtimer;  /* Alarm timer. */
        u_int64_t       p_runtime;      /* Real time in microsec. */
        struct  timeval p_switchtime;   /* When last scheduled */
        u_quad_t p_uticks;              /* Statclock hits in user mode. */
        u_quad_t p_sticks;              /* Statclock hits in system mode. */
        u_quad_t p_iticks;              /* Statclock hits processing intr. */

        int     p_traceflag;            /* Kernel trace points. */
        struct  vnode *p_tracep;        /* Trace to vnode. */

        int     p_siglist;              /* Signals arrived but not delivered. */

        struct  vnode *p_textvp;        /* Vnode of executable. */

        char    p_lock;                 /* Process lock (prevent swap) count. */
        char    p_oncpu;                /* Which cpu we are on */
        char    p_lastcpu;              /* Last cpu we were on */
        char    p_pad2;                 /* alignment */

        short   p_locks;                /* DEBUG: lockmgr count of held locks */
        short   p_simple_locks;         /* DEBUG: count of held simple locks */
        unsigned int    p_stops;        /* procfs event bitmask */
        unsigned int    p_stype;        /* procfs stop event type */
        char    p_step;                 /* procfs stop *once* flag */
        unsigned char   p_pfsflags;     /* procfs flags */
        char    p_pad3[2];              /* padding for alignment */
        register_t p_retval[2];         /* syscall aux returns */
        struct  sigiolst p_sigiolst;    /* list of sigio sources */
#ifdef COMPAT_LINUX_THREADS
        int     p_sigparent;            /* signal to parent on exit */
        sigset_t p_oldsigmask;          /* saved mask from before sigpause */
        int     p_sig;                  /* for core dump/debugger XXX */
        u_long  p_code;                 /* for core dump/debugger XXX */
#endif /* COMPAT_LINUX_THREADS */
/* End area that is zeroed on creation. */
#define p_endzero       p_startcopy

/* The following fields are all copied upon creation in fork. */
#ifndef COMPAT_LINUX_THREADS
#define p_startcopy     p_sigmask
#else
#define p_startcopy     p_procsig
#endif /* COMPAT_LINUX_THREADS */

#ifdef COMPAT_LINUX_THREADS
        struct  procsig *p_procsig;
#define p_sigignore p_procsig->ps_sigignore
#define p_sigcatch p_procsig->ps_sigcatch
#endif /* COMPAT_LINUX_THREADS */
        sigset_t p_sigmask;     /* Current signal mask. */
#ifndef COMPAT_LINUX_THREADS
        sigset_t p_sigignore;   /* Signals being ignored. */
        sigset_t p_sigcatch;    /* Signals being caught by user. */

#endif /* COMPAT_LINUX_THREADS */
        u_char  p_priority;     /* Process priority. */
        u_char  p_usrpri;       /* User-priority based on p_cpu and p_nice. */
        char    p_nice;         /* Process "nice" value. */
        char    p_comm[MAXCOMLEN+1];

        struct  pgrp *p_pgrp;   /* Pointer to process group. */

        struct  sysentvec *p_sysent; /* System call dispatch information. */

        struct  rtprio p_rtprio;        /* Realtime priority. */
/* End area that is copied on creation. */
#define p_endcopy       p_addr
        struct  user *p_addr;   /* Kernel virtual addr of u-area (PROC ONLY). */
        struct  mdproc p_md;    /* Any machine-dependent fields. */

        u_short p_xstat;        /* Exit status for wait; also stop signal. */
        u_short p_acflag;       /* Accounting flags. */
        struct  rusage *p_ru;   /* Exit information. XXX */

        int     p_nthreads;     /* number of threads (only in leader) */
        void    *p_aioinfo;     /* ASYNC I/O info */
        int     p_wakeup;       /* thread id */
        struct proc *p_peers;   
        struct proc *p_leader;
        struct pasleep p_asleep; /* used by asleep()/await()    */
};

struct  pcred {
        struct  ucred *pc_ucred;        /* Current credentials. */
        uid_t   p_ruid;                 /* Real user id. */
        uid_t   p_svuid;                /* Saved effective user id. */
        gid_t   p_rgid;                 /* Real group id. */
        gid_t   p_svgid;                /* Saved effective group id. */
        int     p_refcnt;               /* Number of references. */
};


#endif
